# 
# Hexmap Creation Script C4D R14/R15 Python
# http://www.mikoweb.eu
# miko@mikoweb.de
#
# 0.01 Initial version 10/4/2013
#
#
#                1
#               * * 
#             *     *
#           *    *    *
#        2*    * 7 *    *0
#         *  *8     6*  *
#         *  *       *  *   Indices of a (embossed) hex tile
#         *  *9    11*  *
#        3*    * 10*    *5
#           *    *    *
#             *     *
#               * *
#                4


import math
import array
import os
import c4d
from c4d import gui, utils, bitmaps

# -----------------------------------------------------------
# Constants -------------------------------------------------
HEX_SIZE         = 10              # Nominal size of a hex tile
HEX_INNER        = 0.80 * HEX_SIZE # X-Z Offset of inner hex
HEX_INNER_OFF    = 0.05 * HEX_SIZE # Height offset of inner hex
HEX_XZ_HEIGHT    = 2 * HEX_SIZE    # XZ-plane
HEX_XZ_WIDTH     = math.sqrt(3.0) / 2.0 * HEX_XZ_HEIGHT # XZ-plane

HEX_TYPES        = { 'Water':0, 'Fields':1, 'Hills':2, 'Mountains':3 } # dict
HEX_DIRS         = { 'Left':0, 'LeftUp':1, 'RightUp':2, 'Right':3, 'RightDown':4, 'LeftDown':5 } # dict

# Y position (height) per tile type
HEX_Y_HEIGHT     = [ 0.0, # Water
                     5.0, # Fields
                    10.0, # Hills
                    15.0] # Mountains

# Vertex indices in different directions of the tile
HEX_DIR_VERTS    = [[2,3], # Left
                    [1,2], # LeftUp
                    [0,1], # RightUp
                    [5,0], # Right
                    [4,5], # RightDown
                    [3,4]] # LeftDown

# Just setting UVs so that colors are taken from material's color bitmap
HEX_UVWS         = [[ 7.0/63.0,0.5], # Water
                    [23.0/63.0,0.5], # Fields
                    [39.0/63.0,0.5], # Hills
                    [55.0/63.0,0.5]] # Mountains

# Indices of vertices that build the polys. See index scheme above 
HEX_POLY_INDICES = [[ 6, 8, 7],[ 8, 6,11],[11, 9, 8],[ 9,11,10], 
                    [ 6, 7, 0],[ 0, 7, 1],[ 7, 8, 1],[ 1, 8, 2],
                    [ 8, 9, 2],[ 2, 9, 3],[ 9,10, 3],[ 3,10, 4],
                    [10,11, 4],[ 4,11, 5],[11, 6, 5],[ 5, 6, 0]]

HEX_NO_INDEX     = -1

# Global Vars -----------------------------------------------
gl_MapHexTypes = 0 # To be read in from file (HEX_TYPES)
gl_MapSize     = 0 # To be read in from file
gl_SideWalls   = 0 # Calculated number of side walls for this map


# -----------------------------------------------------------
# Create an embossed hex field (12 verts, 16 polys!) --------
# returns: nothing (adds to polyobj)                 --------
def create_hexfield_bossed (polyobj, vCenterPos, x, z):

    aVertexPos    = [[0,0],[0,0],[0,0],[0,0],[0,0],[0,0],
                     [0,0],[0,0],[0,0],[0,0],[0,0],[0,0]]
                     
    fieldcount = z * gl_MapSize + x
      
    # Offset of indices
    indexVertexOffset = fieldcount * 12
    indexPolyOffset   = fieldcount * 16 
    
    # Calculate vertex positions of hex (thanks http://www.redblobgames.com)
    for i in range(0, 6): # Outer vertices
        angle = 2 * math.pi / 6.0 * (i + 0.5) 
        aVertexPos[i][0] = vCenterPos.x + HEX_SIZE * math.cos(angle)
        aVertexPos[i][1] = vCenterPos.z + HEX_SIZE * math.sin(angle)
        aVertex = c4d.Vector(aVertexPos[i][0], vCenterPos.y, aVertexPos[i][1])
        polyobj.SetPoint(indexVertexOffset + i, aVertex)
        
    for i in range(6, 12): # Inner vertices
        angle = 2 * math.pi / 6.0 * (i -6 + 0.5) 
        aVertexPos[i][0] = vCenterPos.x + HEX_INNER * math.cos(angle)
        aVertexPos[i][1] = vCenterPos.z + HEX_INNER * math.sin(angle)
        aVertex = c4d.Vector(aVertexPos[i][0], vCenterPos.y+HEX_INNER_OFF, aVertexPos[i][1])
        polyobj.SetPoint(indexVertexOffset + i, aVertex)
    
    # Create the corresponding polygons with vertex indices
    for i in range(0, 16):
        aPoly = c4d.CPolygon(indexVertexOffset + HEX_POLY_INDICES[i][0],
                             indexVertexOffset + HEX_POLY_INDICES[i][1],
                             indexVertexOffset + HEX_POLY_INDICES[i][2])
        polyobj.SetPolygon(indexPolyOffset + i, aPoly)
        
    return

# -----------------------------------------------------------
# Create UVW for an embossed hex field               --------
# returns: nothing (adds to polyobj)                 --------
def create_hexfield_bossed_UVW (polyobj, x, z):

    uvtag = polyobj.GetTag(c4d.Tuvw)
    if uvtag == None:
        gui.MessageDialog("Hexmap Creation Script: Could not find UVW Tag! Aborting UVW-Mapping.") 
        return;
    
    dummyUVempty = c4d.Vector(0.0 , 0.0, 0.0)
    
    fieldcount = z * gl_MapSize + x
    polyindex = fieldcount * 16
    fieldtype = gl_MapHexTypes[x][z]
    
    for i in range(0, 16): # 16 poly per hex tile                
                
        indexP1 = HEX_POLY_INDICES[i][0] # Point indices for this poly
        indexP2 = HEX_POLY_INDICES[i][1]
        indexP3 = HEX_POLY_INDICES[i][2]
                
        UV1 = c4d.Vector(HEX_UVWS[fieldtype][0], HEX_UVWS[fieldtype][1], 0.0)
        UV2 = c4d.Vector(HEX_UVWS[fieldtype][0], HEX_UVWS[fieldtype][1], 0.0)
        UV3 = c4d.Vector(HEX_UVWS[fieldtype][0], HEX_UVWS[fieldtype][1], 0.0)               
                
        uvtag.SetSlow(polyindex+i, UV1, UV2, UV3, dummyUVempty)
     
    return;

# -----------------------------------------------------------
# Create a side wall for a given tile and direction ---------
# returns: nothing (adds to polyobj)                ---------
def create_hexfield_wall (polyobj, x1, z1, x2, z2, wallindex, direction):

    vCenterPos1 = get_tile_centerposition(x1, z1)
    vCenterPos2 = get_tile_centerposition(x2, z2)                        
                        
    indexVertexOffset = gl_MapSize * gl_MapSize * 12 + wallindex * 4
    indexPolyOffset   = gl_MapSize * gl_MapSize * 16 + wallindex * 2
    
    angle1 = 2 * math.pi / 6.0 * ( HEX_DIR_VERTS[direction][0] + 0.5)         
    angle2 = 2 * math.pi / 6.0 * ( HEX_DIR_VERTS[direction][1] + 0.5)         
    
    aVertexPos = [[0,0],[0,0]]                
    aVertexPos[0][0] = vCenterPos1.x + HEX_SIZE * math.cos(angle1)
    aVertexPos[0][1] = vCenterPos1.z + HEX_SIZE * math.sin(angle1)
    aVertexPos[1][0] = vCenterPos1.x + HEX_SIZE * math.cos(angle2)
    aVertexPos[1][1] = vCenterPos1.z + HEX_SIZE * math.sin(angle2)
    
    aVertex = c4d.Vector(aVertexPos[0][0], vCenterPos1.y, aVertexPos[0][1])
    polyobj.SetPoint(indexVertexOffset + 0, aVertex)
    aVertex = c4d.Vector(aVertexPos[1][0], vCenterPos1.y, aVertexPos[1][1])
    polyobj.SetPoint(indexVertexOffset + 1, aVertex)
    aVertex = c4d.Vector(aVertexPos[0][0], vCenterPos2.y, aVertexPos[0][1])
    polyobj.SetPoint(indexVertexOffset + 2, aVertex)
    aVertex = c4d.Vector(aVertexPos[1][0], vCenterPos2.y, aVertexPos[1][1])
    polyobj.SetPoint(indexVertexOffset + 3, aVertex)
    
    aPoly = c4d.CPolygon(indexVertexOffset + 0, # Wall poly 1
                         indexVertexOffset + 1,
                         indexVertexOffset + 2)
    polyobj.SetPolygon(indexPolyOffset + 0, aPoly)
    
    aPoly = c4d.CPolygon(indexVertexOffset + 2, # Wall poly 2
                         indexVertexOffset + 1,
                         indexVertexOffset + 3)
    polyobj.SetPolygon(indexPolyOffset + 1, aPoly)
    
    return

# -----------------------------------------------------------
# Create UVW for a side wall                         --------
# returns: nothing (adds to polyobj)                 --------
def create_hexfield_wall_UVW (polyobj, x, z, wallindex, direction):

    uvtag = polyobj.GetTag(c4d.Tuvw)
    if uvtag == None:
        gui.MessageDialog("Hexmap Creation Script: Could not find UVW Tag! Aborting UVW-Mapping.") 
        return;
    
    dummyUVempty = c4d.Vector(0.0 , 0.0, 0.0)  
    
    # Assign UVW for the side wall polys
    polyindex = gl_MapSize * gl_MapSize * 16 + wallindex * 2
    fieldtype = gl_MapHexTypes[x][z]
    
    for i in range(polyindex, polyindex+2): # 2 poly per side wall
            
        UV1 = c4d.Vector(HEX_UVWS[fieldtype][0], HEX_UVWS[fieldtype][1], 0.0)
        UV2 = c4d.Vector(HEX_UVWS[fieldtype][0], HEX_UVWS[fieldtype][1], 0.0)
        UV3 = c4d.Vector(HEX_UVWS[fieldtype][0], HEX_UVWS[fieldtype][1], 0.0)
            
        uvtag.SetSlow(i, UV1, UV2, UV3, dummyUVempty)
     
    return;

# -----------------------------------------------------------
# Add a hextile center position to data file         --------
# returns: nothing                                   --------
def add_hextile_position_to_file(indexX, indexZ, vCenterPos):

    sFilePath = doc.GetDocumentPath() + "\\Data_HexPos.txt"
    
    fp = open(sFilePath, "a+")
    fp.write(str(indexX)+";"+str(indexZ)+";")
    fp.write(str(vCenterPos.x)+";"+str(vCenterPos.y)+";"+str(vCenterPos.z)+"\n")
    fp.close()
        
    return

# -----------------------------------------------------------
# Deletes the hextile center position data file      --------
# returns: nothing                                   --------
def delete_hextile_position_file():

    sFilePath = doc.GetDocumentPath() + "\\Data_HexPos.txt"
    if (os.path.exists(sFilePath) == True):
        os.remove (sFilePath);
    return

# -----------------------------------------------------------
# Check if there is a save path available            --------
# returns: true/false                                --------
def check_for_save_path():

    sPath = doc.GetDocumentPath()
    if (sPath == ""):
        return False
    else:
        return True

# -----------------------------------------------------------
# Count how many tile side walls we need             --------
# returns: number of needed side walls               --------
def get_sidewalls_number():
    
    # Step through all tiles and mark a wall as needed if
    # adjacent tile is of different height and is lower
    count = 0
    for z in range(0, gl_MapSize):
        for x in range(0, gl_MapSize):
            for direction in range(0,6):
                (retX, retZ) = get_neighbor_tile(x,z,direction)   
                if (is_tileindex_valid(retX,retZ)):
                    if (gl_MapHexTypes[x][z] > gl_MapHexTypes[retX][retZ]):
                        count += 1;

    return count

# -----------------------------------------------------------
# Check if there is a map definition file to read in --------
# returns: true/false                                --------
def check_for_mapdef_file():

    sFilePath = doc.GetDocumentPath() + "\\mymap_xz.txt"
    if (os.path.exists(sFilePath) == False):
        return False
    else:
        return True

# -----------------------------------------------------------
# Check if there is a bitmap for the shader          --------
# returns: true/false                                --------
def check_for_shaderbitmap_file():

    sFilePath = doc.GetDocumentPath() + "\\mymap.bmp"
    if (os.path.exists(sFilePath) == False):
        return False
    else:
        return True
    
# -----------------------------------------------------------
# Read in hex tile data from map def file            --------
# returns: nothing (sets gl_MapSize,gl_MapHexTypes)  --------
def read_in_mapdef_file():

    global gl_MapSize
    global gl_MapHexTypes
    
    sFilePath = doc.GetDocumentPath() + "\\mymap_xz.txt"
    
    fp = open(sFilePath, "r")
    for i in range(0, 4):
        fp.readline() # get rid of header lines
    
    # Read in map size info
    aLine = fp.readline() # Example: "40,40" 
    sizeList = aLine.split(",")
    gl_MapSize = int(sizeList[0])
    gl_MapHexTypes = [[0 for i in range(gl_MapSize)] for j in range(gl_MapSize)]
    
    # Read in hex tile types (not positions; only indices and types)
    for i in range(0, gl_MapSize * gl_MapSize):
         aLine = fp.readline() # Example: "0,1,4.330127,0.000000,7.500000,Fields"
         dataList = aLine.split(",")
         indexX  = int(dataList[0])
         indexZ  = int(dataList[1])
         hexYpos = int(HEX_TYPES[dataList[5].strip()])
         gl_MapHexTypes[indexX][indexZ] = hexYpos
         
    fp.close()
        
    return

# -----------------------------------------------------------
# Create a new material and assign to object         --------
# returns: nothing (adds to polyobj)                 --------
def create_and_assign_material(polyobj, sName):
    
    sFilePath = doc.GetDocumentPath() + "\\mymap.bmp"
    
    # Remove existing if found
    old_mat = doc.SearchMaterial(sName)
    if (old_mat):
        old_mat.Remove()
    
    # Create and add material    
    mat = c4d.BaseMaterial(c4d.Mmaterial)
    mat.SetName(sName)
    doc.InsertMaterial(mat)
    
    # Define bitmap shader for the material
    shader = c4d.BaseList2D(c4d.Xbitmap)
    shader[c4d.BITMAPSHADER_FILENAME] = sFilePath
    mat.InsertShader(shader)
    mat[c4d.MATERIAL_COLOR_SHADER] = shader
    
    # Create texture tag and add to the material
    ttag = c4d.BaseTag(c4d.Ttexture)
    ttag[c4d.TEXTURETAG_MATERIAL] = mat
    ttag[c4d.TEXTURETAG_PROJECTION] = c4d.TEXTURETAG_PROJECTION_UVW
    ttag.Message(c4d.MSG_UPDATE)
    polyobj.InsertTag(ttag)
    
    # Create uvw tag and add to the material
    uvtag = c4d.VariableTag(c4d.Tuvw, gl_MapSize * gl_MapSize * 16 + gl_SideWalls * 2) # polycount
    
    uvtag.Message(c4d.MSG_UPDATE)
    polyobj.InsertTag(uvtag)
    
    mat.Update(True,True)
    mat.Message(c4d.MSG_UPDATE) 
    
    return

# -----------------------------------------------------------
# Insert new object into hierarchy. Replace if present ------
# returns: nothing                                     ------
def replace_object_with_name(polyobj, sName):
    
    # Remove existing if found
    old_polyobject = doc.SearchObject(sName)
    if (old_polyobject):
        old_polyobject.Remove()
    
    # Assign new name
    polyobj.SetName(sName)
         
    # Insert new object into hierarchy
    doc.InsertObject(polyobj,None,None)
    
    return

# -----------------------------------------------------------
# Returns map index from adjacent tile               --------
# returns: tile index                                --------
def get_neighbor_tile(indexX,indexZ,direction):    

    retX = indexX
    retZ = indexZ
    
    if (direction ==  HEX_DIRS['Left']):
        retX -= 1
            
    if (direction ==  HEX_DIRS['LeftUp']):
        retZ += 1
        retX -= (1-indexZ%2)
        
    if (direction ==  HEX_DIRS['RightUp']):
        retZ += 1
        retX += indexZ%2
        
    if (direction ==  HEX_DIRS['Right']):
        retX += 1
        
    if (direction ==  HEX_DIRS['RightDown']):
        retZ -= 1
        retX += indexZ%2
        
    if (direction ==  HEX_DIRS['LeftDown']):
        retZ -= 1
        retX -= (1-indexZ%2)

    if (is_tileindex_valid(retX, retZ)==True):
        return (retX, retZ)
    
    return (HEX_NO_INDEX, HEX_NO_INDEX)

# -----------------------------------------------------------
# Checks if map indices are valid                    --------
# returns: true/false                                --------
def is_tileindex_valid(indexX,indexZ):    
    
    return ((indexX > HEX_NO_INDEX) and
            (indexZ > HEX_NO_INDEX) and 
            (indexX < gl_MapSize)   and 
            (indexZ < gl_MapSize))

# -----------------------------------------------------------
# Returns a hextile world center position            --------
# returns: Position vector                           --------
def get_tile_centerposition(x,z):  
    
    odd_row_offset = (z % 2) * HEX_XZ_WIDTH / 2.0
    MapHexType     = gl_MapHexTypes[x][z]
    
    vCenterPos =  c4d.Vector(odd_row_offset + x * HEX_XZ_WIDTH,
                             HEX_Y_HEIGHT[MapHexType],
                             3.0/4.0 * HEX_XZ_HEIGHT * z)
                             
    return (vCenterPos)
    
# -----------------------------------------------------------
# MAIN
# -----------------------------------------------------------
def main():
    
    global gl_SideWalls
    
    print("Hexmap Creation Script: STARTED");
     
    if check_for_save_path() == False:
        gui.MessageDialog("Hexmap Creation Script: No path to save to. Create a model file! Aborting script.")
        return
    
    if check_for_mapdef_file() == False:
        gui.MessageDialog("Hexmap Creation Script: No map def file 'mymap_xz.txt' to read in! Aborting script.") 
        return;
    
    if check_for_shaderbitmap_file() == False:
        gui.MessageDialog("Hexmap Creation Script: No bitmap file 'mymap.bmp' to read in! Aborting script.") 
        return;
       
    read_in_mapdef_file() # Read in map config from mymap_xz.txt
    delete_hextile_position_file() # Delete info file we will create again later
    
    gl_SideWalls = get_sidewalls_number() # How many side walls do we need?
    print('Needed tile side walls: '+str(gl_SideWalls))
    
    mesh_vertexcount = gl_MapSize * gl_MapSize * 12 + gl_SideWalls * 4
    mesh_polycount   = gl_MapSize * gl_MapSize * 16 + gl_SideWalls * 2
    
    print('Calculated mesh_vertexcount: '+str(mesh_vertexcount))
    print('Calculated mesh_polycount: '+str(mesh_polycount))
    
    # Create the polyobject to be filled    
    polyobj = c4d.PolygonObject(mesh_vertexcount, mesh_polycount)
    create_and_assign_material(polyobj, "HexMat")    
    replace_object_with_name(polyobj,"HexagonMap")
      
    # Calculate hex positions and create fields
    for z in range(0, gl_MapSize):
        for x in range(0, gl_MapSize):            
            
            MapHexType = gl_MapHexTypes[x][z]
            vCenterPos =  get_tile_centerposition(x, z)
            add_hextile_position_to_file(x, z, vCenterPos)     
            create_hexfield_bossed(polyobj, vCenterPos, x, z)
            create_hexfield_bossed_UVW(polyobj, x, z)   
    
    # Calculate and create the tile walls (vertices+polys)
    wallindex = 0
    for z in range(0, gl_MapSize):
        for x in range(0, gl_MapSize):
            for direction in range(0,6):
                (retX, retZ) = get_neighbor_tile(x,z,direction)   
                if (is_tileindex_valid(retX,retZ)):
                    if (gl_MapHexTypes[x][z] > gl_MapHexTypes[retX][retZ]): # Higher terrain lvl?
                        create_hexfield_wall (polyobj, x, z, retX, retZ, wallindex, direction)
                        create_hexfield_wall_UVW (polyobj, x, z, wallindex, direction)
                        wallindex += 1    
    
    polyobj.SetPhong(True, 1, utils.Rad(10.0)) 
   
    polyobj.GetTag(c4d.Tuvw).Message(c4d.MSG_UPDATE)
    polyobj.Message(c4d.MSG_UPDATE)
    
    c4d.EventAdd()
    
    print("Hexmap Creation Script: FINISHED");
    
    return

# -----------------------------------------------------------
        
if __name__=='__main__':
    main()
